/*
 * client_config.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "anim.hpp"
#include "binary_io/binary_io.hpp"
#include "control.hpp"
#include "client_config.hpp"
#include "misc/my_exceptions.hpp"
#include "platform/graphic.hpp"
#include "potion_renderer.hpp"
#include "skull_renderer.hpp"

ClientConfig::ClientConfig(istream &str)
{
	// NOTE: we load all the gfx config, overlay offsets etc from the network. This probably
	// means that config needs to be the same on both ends (eg changing the gfx sizes would
	// mess up the overlay offsets) ... Coming up with a better method needs some thought
	// though.
	
	InitOverlayOffsets(str);
	
	int n;
	n = BinReadLong(str);
	graphics.resize(n);
	for (int i=0; i<n; ++i) {
		graphics[i] = CreateGraphicFromStream(str);
	}

	n = BinReadLong(str);
	anims.resize(n);
	for (int i=0; i<n; ++i) {
		anims[i] = new Anim(str, *this);
	}

	n = BinReadLong(str);
	overlays.resize(n);
	for (int i=0; i<n; ++i) {
		overlays[i] = new Overlay(str, *this);
	}

	n = BinReadLong(str);
	if (n != NUM_STANDARD_CONTROLS) throw UnexpectedError("standard controls do not match");
	n = BinReadLong(str);
	controls.resize(n+NUM_STANDARD_CONTROLS);
	for (int i=0; i<NUM_STANDARD_CONTROLS; ++i) {
		controls[i] = g_standard_controls[i];
	}
	for (int i=NUM_STANDARD_CONTROLS; i<n+NUM_STANDARD_CONTROLS; ++i) {
		controls[i] = new Control(str, *this, i+1);
	}

	potion_renderer.reset(new PotionRenderer(str, *this));
	skull_renderer.reset(new SkullRenderer(str, *this));

	menu_centre = getGraphic(BinReadLong(str));
	menu_empty = getGraphic(BinReadLong(str));
	menu_highlight = getGraphic(BinReadLong(str));
}

ClientConfig::~ClientConfig()
{
	for (int i=0; i<controls.size(); ++i) delete controls[i];
	for (int i=0; i<overlays.size(); ++i) delete overlays[i];
	for (int i=0; i<anims.size(); ++i) delete anims[i];
	for (int i=0; i<graphics.size(); ++i) DeleteGraphic(graphics[i]);
}

Graphic * ClientConfig::getGraphic(int id) const
{
	if (id < 1 || id > graphics.size()) return 0;
	else return graphics[id-1];
}

Anim * ClientConfig::getAnim(int id) const
{
	if (id < 1 || id > anims.size()) return 0;
	else return anims[id-1];
}

Overlay * ClientConfig::getOverlay(int id) const
{
	if (id < 1 || id > overlays.size()) return 0;
	else return overlays[id-1];
}

Control * ClientConfig::getControl(int id) const
{
	if (id < 1 || id > controls.size()) return 0;
	else return controls[id-1];
}

