/*
 * lockable.hpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 * Lockable: Base class for openable/trappable/lockable tiles (ie doors and chests).
 *
 */

#ifndef LOCKABLE_HPP
#define LOCKABLE_HPP

#include "tile.hpp"

#include "boost/shared_ptr.hpp"
using namespace boost;

class Creature;
class DungeonMap;
class MapCoord;
class Trap;

class Lockable : public Tile {
public:
	Lockable() : closed(true), can_trap(true), lock(0) { }

	void setOpenInitially() { closed = false; }
	void setLockChance(int percent) { lock = -percent; }
	void setCannotTrap() { can_trap = false; }

	// Override canActivateHere, onActivate, onOpen and onClose to deal with opening and
	// closing
	virtual bool canActivateHere() const { return false; }
	virtual void onActivate(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature>,
							ActivateType, bool success = true);
	virtual void onClose(DungeonMap &, const MapCoord &, shared_ptr<Creature>);
	virtual void onOpen(DungeonMap &, const MapCoord &, shared_ptr<Creature>);
	
	bool isClosed() const { return closed; }
	
	// Locks
	bool isLocked() const { return lock > 0; }
	void generateLock(int nkeys);  // If "lock" is negative then this generates a lock.

	// Traps
	// canTrap: true if this Lockable can have traps applied to it
	// setTrap: Set a trap which will be triggered when the door/chest is opened. If a trap
	// is already present, then the existing trap will be set off.
	// generateTrap: like setTrap but called during initial "pretrapped chests" generation.
	// (returns true if a trap was generated.)
	// onHit: Override this to deal with setting off traps when a door/chest is hit.
	bool canTrap() const { return can_trap; }
	void setTrap(DungeonMap &dmap, const MapCoord &mc, shared_ptr<Creature> cr,
				 shared_ptr<Trap> newtrap);
	virtual bool generateTrap(DungeonMap &, const MapCoord &) { return false; }
	virtual void onHit(DungeonMap &, const MapCoord &, shared_ptr<Creature>);

protected:
	// "open", "close" are implemented by subclasses, and handle the actual opening
	// and closing.
	virtual void open(DungeonMap &, const MapCoord &) = 0;
	virtual void close(DungeonMap &, const MapCoord &) = 0;	

private:
	bool doOpen(DungeonMap &, const MapCoord &, shared_ptr<Creature>, ActivateType);
	bool checkUnlock(shared_ptr<Creature>) const;
	void activateTraps(DungeonMap &, const MapCoord &, shared_ptr<Creature>);
	void disarmTraps(DungeonMap &, const MapCoord &, shared_ptr<Creature>);
	
private:
	bool closed;
	bool can_trap;
	int lock; // key number, or zero for unlocked
	// NOTE: a negative value for lock means that lock has not been generated yet, and there
	// is a (-lock) percent chance of the door being locked.
	shared_ptr<Trap> trap;
};

#endif
