/*
 * mini_map.cpp
 *
 * Copyright (c) Stephen Thompson, 2007.
 * Licenced for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "map_support.hpp"
#include "mini_map.hpp"
#include "platform/display.hpp"

#include <algorithm>
using namespace std;

namespace {
	int col[][3] = {
		{ 0,0,0 },  // COL_UNMAPPED
		{ 136,136,85 }, // COL_FLOOR
		{ 68,68,34 }, // COL_WALL
		{ 255,0,0 },     // highlight
		{ 255,255,255 }, // highlight
		{ 255,0,0 },     // highlight
		{ 0,0,0 }        // highlight
	};

	const int FIRST_HIGHLIGHT = COL_WALL+1;
	const int NUM_HIGHLIGHTS = 4;
}

void MiniMap::create(int w, int h)
{
	if (width != 0 || w <= 0 || h <= 0) return;
	width = w;
	height = h;
	data.resize(width*height);
}	

MiniMapColour MiniMap::getColour(const MapCoord &mc) const
{
	if (mc.getX() < 0 || mc.getX() >= width || mc.getY() < 0 || mc.getY() >= height) {
		return COL_UNMAPPED;
	} else {
		return data[mc.getY()*width + mc.getX()];
	}
}

void MiniMap::setHighlight(const MapCoord &mc, int id)
{
	// The strategy is to store highlights into a separate map, rather than
	// simply setting colours directly in the "data" array. This simplifies
	// setHighlight and setColour, but it does complicate the draw routines
	// slightly. 
	
	// Erase the old highlight first (if any)
	map<int,Highlight>::iterator it = highlights.find(id);
	if (it != highlights.end()) {
		highlights.erase(it);
	}

	if (!mc.isNull() && mc.getX() >= 0 && mc.getX() < width
	&& mc.getY() >= 0 && mc.getY() < height) {
		// Set the new highlight.
		Highlight h;
		h.pos = mc;
		h.old_colour = data[mc.getY()*width+mc.getX()];
		highlights.insert(make_pair(id, h));
	}
}

void MiniMap::setColour(const MapCoord &mc, MiniMapColour col)
{
	if (mc.getX() < 0 || mc.getX() >= width || mc.getY() < 0 || mc.getY() >= height) return;
	data[mc.getY()*width + mc.getX()] = col;
}

void MiniMap::wipe()
{
	fill(data.begin(), data.end(), COL_UNMAPPED);
}

MiniMap::Highlighter::Highlighter(const MiniMap &m, int t)
	: mini_map(m)
{
	const MiniMapColour col_highlight = MiniMapColour((t / CfgInt("mini_map_flash_time")) % NUM_HIGHLIGHTS + FIRST_HIGHLIGHT);
	
	for (map<int,MiniMap::Highlight>::const_iterator it = mini_map.highlights.begin();
	it != mini_map.highlights.end(); ++it) {
		mini_map.data[it->second.pos.getY()*mini_map.width + it->second.pos.getX()]
			= col_highlight;
	}
}

MiniMap::Highlighter::~Highlighter()
{
	// Reset highlights to their proper map colour.
	for (map<int,MiniMap::Highlight>::const_iterator it = mini_map.highlights.begin();
	it != mini_map.highlights.end(); ++it) {
		mini_map.data[it->second.pos.getY()*mini_map.width + it->second.pos.getX()]
		    = it->second.old_colour;
	}
}

void MiniMap::draw(Display &dpy, int bl_x, int bl_y, int scale, int t) const
{
	// Set highlights to their proper colour.
	Highlighter h(*this, t);
	
	// Draw the mini map using "scale" pixels for each square.
	const int dw = dpy.getWidth();
	const int dh = dpy.getHeight();
	if (bl_x >= dw) return;
	if (bl_y >= dh) return;
	if (scale < 1) return;

	// Clip to screen
	// -- work out ranges [startx,finishx) and [starty,finishy) to draw.
	
	int startx = 0;
	if (bl_x < 0) {
		startx = (-bl_x + scale-1) / scale;
		bl_x = 0;
	}

	int starty = 0;
	if (bl_y < 0) {
		starty = (-bl_y + scale-1) / scale;
		bl_y = 0;
	}

	int finishx = width;
	const int xpixels = (finishx-startx)*scale;
	if (bl_x + xpixels > dw) {
		finishx -= (bl_x + xpixels - dw + scale-1) / scale;
	}

	int finishy = height;
	const int ypixels = (finishy-starty)*scale;
	if (bl_y + ypixels > dh) {
		finishy -= (bl_y + ypixels - dh + scale-1) / scale;
	}

	// Check if entire map was clipped
	if (finishx <= startx) return;
	if (finishy <= starty) return;

	// Do the drawing
	dpy.beginPixels();
	for (int x=startx; x<finishx; ++x) {
		for (int y=starty; y<finishy; ++y) {
			const int c = data[y*width+x];
			const int r = col[c][0];
			const int g = col[c][1];
			const int b = col[c][2];

			const int basex = (x-startx)*scale + bl_x;
			const int basey = (y-starty)*scale + bl_y;
			for (int j=0; j<scale; ++j) {
				for (int i=0; i<scale; ++i) {
					dpy.drawPixel(basex+i, basey+j, r, g, b);
				}
			}
		}
	}
	dpy.endPixels();
}


void MiniMap::drawOnePointFiveScale(Display &dpy, int bl_x, int bl_y, int t) const
{
	Highlighter h(*this, t);
	
	// Draw the mini map using 3 pixels per 2 squares
	// (useful for small displays, eg PocketPC)
	
	const int dw = dpy.getWidth();
	const int dh = dpy.getHeight();
	if (bl_x >= dw) return;
	if (bl_y >= dh) return;

	// Compute subset of the map to draw (for clipping).
	// Map coords drawn are in range [startx,finishx) and [starty,finishy).
	// (Screen coords always start from bl_x and bl_y.)

	// Starting position: 0 unless bl_x is negative, in which case we add 2 squares
	// for each 3 pixels (or part thereof), and reset bl_x to zero. 
	int startx = 0;
	if (bl_x < 0) {
		startx = (-bl_x + 2) / 3 * 2;
		bl_x = 0;
	}

	// same for y
	int starty = 0;
	if (bl_y < 0) {
		starty = (-bl_y + 2) / 3 * 2;
		bl_y = 0;
	}

	// Pixels needed to draw [startx,finishx): (finishx-startx)*1.5.
	// If this is too many: deduct 2 from finishx for each excess 3 pixels (or part thereof).
	int finishx = width%2==0 ? width : width+1;  // must be even
	const int xpixels = (finishx-startx)*3/2;
	if (bl_x + xpixels > dw) {
		finishx -= (bl_x + xpixels - dw + 2) / 3 * 2;
	}

	// Same for y
	int finishy = height%2==0 ? height : height+1;
	const int ypixels = (finishy-starty)*3/2;
	if (bl_y + ypixels > dh) {
		finishy -= (bl_y + ypixels - dh + 2) / 3 * 2;
	}

	// check whether entire map was clipped...
	if (finishx <= startx) return;
	if (finishy <= starty) return;

	// start the drawing
	dpy.beginPixels();

	// We do this by looping through 2x2 blocks.
	ASSERT(startx%2==0);
	ASSERT(starty%2==0);
	ASSERT(finishx%2==0);
	ASSERT(finishy%2==0);
	for (int x=startx; x<finishx; x += 2) {
		for (int y=starty; y<finishy; y += 2) {

			// Get the four colours.
			// The range checks are necessary because we might have rounded up the map
			// size to make it even.
			const int bl = data[y*width+x];
			const int br = (x==width-1) ? COL_UNMAPPED : data[y*width+x+1];
			const int tl = (y==height-1) ? COL_UNMAPPED : data[(y+1)*width+x];
			const int tr = (x==width-1 || y==height-1) ? COL_UNMAPPED : data[(y+1)*width+x+1];

			// Compute intermediate colours.
			int output[3][3][3];  // x, y, rgb index
			for (int j=0; j<3; ++j) {
				output[0][0][j] = col[bl][j];
				output[2][0][j] = col[br][j];
				output[0][2][j] = col[tl][j];
				output[2][2][j] = col[tr][j];

				// interpolate to get 'edge centre' colours
				output[1][0][j] = (output[0][0][j] + output[2][0][j]) / 2;
				output[1][2][j] = (output[0][2][j] + output[2][2][j]) / 2;
				output[0][1][j] = (output[0][0][j] + output[0][2][j]) / 2;
				output[2][1][j] = (output[2][0][j] + output[2][2][j]) / 2;

				// interpolate to get central colour.
				// (it does not matter whether we average [1][0] and [1][2], or
				// [0][1] and [2][1])
				output[1][1][j] = (output[0][1][j] + output[2][1][j]) / 2;
			}

			// Now plot
			const int basex = (x-startx)/2*3 + bl_x;
			const int basey = (y-starty)/2*3 + bl_y;
			for (int i=0; i<3; ++i) {
				for (int j=0; j<3; ++j) {
					dpy.drawPixel(basex+i, basey+j, output[i][j][0], output[i][j][1],
								  output[i][j][2]);
				}
			}
		}
	}

	dpy.endPixels();
}

